@@ -13,7 +13,9 @@ import six
 import shutil
 import pip as pip_mod
 from datetime import datetime
+from concurrent.futures._base import Future
 from localstack import constants, config
+from localstack.utils.analytics.profiler import log_duration
 
 # set up logger
 LOG = logging.getLogger(os.path.basename(__file__))
@@ -42,19 +44,19 @@ PLUGIN_SCOPE_COMMANDS = 'commands'
 
 # maps from API names to list of other API names that they depend on
 API_DEPENDENCIES = {
+    'dynamodb': ['dynamodbstreams'],
     'dynamodbstreams': ['kinesis'],
     'es': ['elasticsearch'],
-    'lambda': ['logs']
+    'lambda': ['logs', 'cloudwatch'],
+    'kinesis': ['dynamodb'],
+    'firehose': ['kinesis']
 }
 # composites define an abstract name like "serverless" that maps to a set of services
 API_COMPOSITES = {
-    'serverless': ['cloudformation', 'iam', 'sts', 'lambda', 'dynamodb', 'apigateway', 's3'],
+    'serverless': ['cloudformation', 'cloudwatch', 'iam', 'sts', 'lambda', 'dynamodb', 'apigateway', 's3'],
     'cognito': ['cognito-idp', 'cognito-identity']
 }
 
-# name of main Docker container
-MAIN_CONTAINER_NAME = 'localstack_main'
-
 # environment variable that indicates that we're executing in
 # the context of the script that starts the Docker container
 ENV_SCRIPT_STARTING_DOCKER = 'LS_SCRIPT_STARTING_DOCKER'
@@ -99,6 +101,7 @@ def run_pip_main(args):
     return pip._internal.main.main(args)
 
 
+@log_duration()
 def load_plugin_from_path(file_path, scope=None):
     if os.path.exists(file_path):
         module = re.sub(r'(^|.+/)([^/]+)/plugins.py', r'\2', file_path)
@@ -123,6 +126,13 @@ def load_plugin_from_path(file_path, scope=None):
                 LOG.warning('Unable to load plugins from file %s: %s' % (file_path, e))
 
 
+def should_load_module(module, scope):
+    if module == 'localstack_ext' and not os.environ.get('LOCALSTACK_API_KEY'):
+        return False
+    return True
+
+
+@log_duration()
 def load_plugins(scope=None):
     scope = scope or PLUGIN_SCOPE_SERVICES
     if PLUGINS_LOADED.get(scope):
@@ -139,6 +149,8 @@ def load_plugins(scope=None):
     search_modules = PLUGIN_MODULES
 
     for module in search_modules:
+        if not should_load_module(module, scope):
+            continue
         file_path = None
         if isinstance(module, six.string_types):
             loader = pkgutil.get_loader(module)
@@ -174,7 +186,7 @@ def docker_container_running(container_name):
 
 
 def get_docker_container_names():
-    cmd = "docker ps --format '{{.Names}}'"
+    cmd = "%s ps --format '{{.Names}}'" % config.DOCKER_CMD
     try:
         output = to_str(run(cmd))
         container_names = re.split(r'\s+', output.strip().replace('\n', ' '))
@@ -185,12 +197,22 @@ def get_docker_container_names():
 
 
 def get_main_container_ip():
-    cmd = "docker inspect -f '{{range .NetworkSettings.Networks}}{{.IPAddress}}{{end}}' %s" % MAIN_CONTAINER_NAME
+    container_name = get_main_container_name()
+    cmd = ("%s inspect -f '{{range .NetworkSettings.Networks}}{{.IPAddress}}{{end}}' %s" %
+        (config.DOCKER_CMD, container_name))
     return run(cmd).strip()
 
 
+def get_main_container_name():
+    cmd = "%s inspect -f '{{ .Name }}' %s" % (config.DOCKER_CMD, config.HOSTNAME)
+    return run(cmd).strip().lstrip('/')
+
+
 def setup_logging():
     # determine and set log level
+    if PLUGINS_LOADED.get('_logging_'):
+        return
+    PLUGINS_LOADED['_logging_'] = True
     log_level = logging.DEBUG if is_debug() else logging.INFO
     logging.basicConfig(level=log_level, format=LOG_FORMAT, datefmt=LOG_DATE_FORMAT)
 
@@ -298,8 +320,10 @@ class PortMappings(object):
             for i in range(port[1] - port[0] + 1):
                 self.add(port[0] + i, mapped[0] + i)
             return
-        if int(port) <= 0:
+        if port is None or int(port) <= 0:
             raise Exception('Unable to add mapping for invalid port: %s' % port)
+        if self.contains(port):
+            return
         for from_range, to_range in self.mappings.items():
             if not self.in_expanded_range(port, from_range):
                 continue
@@ -318,6 +342,11 @@ class PortMappings(object):
 
         return ' '.join([entry(k, v) for k, v in self.mappings.items()])
 
+    def contains(self, port):
+        for from_range, to_range in self.mappings.items():
+            if self.in_range(port, from_range):
+                return True
+
     def in_range(self, port, range):
         return port >= range[0] and port <= range[1]
 
@@ -339,14 +368,29 @@ def get_docker_image_to_start():
     image_name = os.environ.get('IMAGE_NAME')
     if not image_name:
         image_name = constants.DOCKER_IMAGE_NAME
-        if os.environ.get('USE_LIGHT_IMAGE') in constants.TRUE_STRINGS:
-            image_name = constants.DOCKER_IMAGE_NAME_LIGHT
+        if os.environ.get('USE_LIGHT_IMAGE') in constants.FALSE_STRINGS:
+            image_name = constants.DOCKER_IMAGE_NAME_FULL
     return image_name
 
 
+def extract_port_flags(user_flags, port_mappings):
+    regex = r'-p\s+([0-9]+)(\-([0-9]+))?:([0-9]+)(\-([0-9]+))?'
+    matches = re.match('.*%s' % regex, user_flags)
+    start = end = 0
+    if matches:
+        for match in re.findall(regex, user_flags):
+            start = int(match[0])
+            end = int(match[2] or match[0])
+            start_target = int(match[3] or start)
+            end_target = int(match[5] or end)
+            port_mappings.add([start, end], [start_target, end_target])
+        user_flags = re.sub(regex, r'', user_flags)
+    return user_flags
+
+
 def start_infra_in_docker():
 
-    container_name = MAIN_CONTAINER_NAME
+    container_name = config.MAIN_CONTAINER_NAME
 
     if docker_container_running(container_name):
         raise Exception('LocalStack container named "%s" is already running' % container_name)
@@ -359,7 +403,6 @@ def start_infra_in_docker():
     # prepare APIs
     canonicalize_api_names()
 
-    services = os.environ.get('SERVICES', '')
     entrypoint = os.environ.get('ENTRYPOINT', '')
     cmd = os.environ.get('CMD', '')
     user_flags = config.DOCKER_FLAGS
@@ -375,13 +418,8 @@ def start_infra_in_docker():
     port_mappings = PortMappings()
 
     # get port ranges defined via DOCKER_FLAGS (if any)
-    regex = r'.*-p\s+([0-9]+)(\-([0-9]+))?:([0-9]+)(\-[0-9]+)?.*'
-    match = re.match(regex, user_flags)
-    start = end = 0
-    if match:
-        start = int(match.group(1))
-        end = int(match.group(3) or match.group(1))
-        port_mappings.add([start, end])
+    user_flags = extract_port_flags(user_flags, port_mappings)
+    plugin_run_params = extract_port_flags(plugin_run_params, port_mappings)
 
     # construct default port mappings
     if service_ports.get('edge') == 0:
@@ -390,11 +428,6 @@ def start_infra_in_docker():
     for port in service_ports.values():
         port_mappings.add(port)
 
-    if services:
-        port_mappings = PortMappings()
-        for port in set(service_ports.values()):
-            port_mappings.add(port)
-
     env_str = ''
     for env_var in config.CONFIG_ENV_VARS:
         value = os.environ.get(env_var, None)
@@ -493,23 +526,31 @@ class FuncThread(threading.Thread):
         self.params = params
         self.func = func
         self.quiet = quiet
+        self.result_future = Future()
 
     def run(self):
+        result = None
         try:
-            self.func(self.params)
-        except Exception:
+            result = self.func(self.params)
+        except Exception as e:
+            result = e
             if not self.quiet:
-                LOG.warning('Thread run method %s(%s) failed: %s' %
-                    (self.func, self.params, traceback.format_exc()))
+                LOG.warning('Thread run method %s(%s) failed: %s %s' %
+                    (self.func, self.params, e, traceback.format_exc()))
+        finally:
+            try:
+                self.result_future.set_result(result)
+            except Exception:
+                # this can happen as InvalidStateError on shutdown, if the task is already canceled
+                pass
 
     def stop(self, quiet=False):
         if not quiet and not self.quiet:
             LOG.warning('Not implemented: FuncThread.stop(..)')
 
 
-def run(cmd, print_error=True, asynchronous=False, stdin=False,
-        stderr=subprocess.STDOUT, outfile=None, env_vars=None, inherit_cwd=False,
-        inherit_env=True, tty=False):
+def run(cmd, print_error=True, asynchronous=False, stdin=False, stderr=subprocess.STDOUT,
+        outfile=None, env_vars=None, inherit_cwd=False, inherit_env=True, tty=False):
     env_dict = os.environ.copy() if inherit_env else {}
     if env_vars:
         env_dict.update(env_vars)
@@ -569,19 +610,18 @@ def run(cmd, print_error=True, asynchronous=False, stdin=False,
 
 
 def is_mac_os():
-    try:
-        out = to_str(subprocess.check_output('uname -a', shell=True))
-        return 'Darwin' in out
-    except subprocess.CalledProcessError:
-        return False
+    return 'Darwin' in get_uname()
 
 
 def is_linux():
+    return 'Linux' in get_uname()
+
+
+def get_uname():
     try:
-        out = to_str(subprocess.check_output('uname -a', shell=True))
-        return 'Linux' in out
-    except subprocess.CalledProcessError:
-        return False
+        return to_str(subprocess.check_output('uname -a', shell=True))
+    except Exception:
+        return ''
 
 
 def mkdir(folder):
